--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   mt_mqueue_remote
--
-- description: Remote MQ implementation. Exchanges messages between CPU CBs in
-- remote nodes.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2014-2018
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

use work.wishbone_pkg.all;
use work.mt_mqueue_pkg.all;

entity mt_mqueue_remote is
  generic (
    g_CONFIG : t_mt_mqueue_config := c_MT_DEFAULT_MQUEUE_CONFIG);
  port (
    clk_i        : in  std_logic;
    rst_n_i      : in  std_logic;
    si_slave_i   : in  t_wishbone_slave_in;
    si_slave_o   : out t_wishbone_slave_out;
    --  From OUT rmq
    src_o        : out t_mt_stream_source_out_array(0 to g_CONFIG.slot_count-1);
    src_i        : in  t_mt_stream_source_in_array(0 to g_CONFIG.slot_count-1);
    --  To IN rmq
    snk_o        : out t_mt_stream_sink_out_array(0 to g_CONFIG.slot_count-1);
    snk_i        : in  t_mt_stream_sink_in_array(0 to g_CONFIG.slot_count-1);
    rmq_status_o : out std_logic_vector(g_CONFIG.slot_count-1 downto 0));
end mt_mqueue_remote;

architecture arch of mt_mqueue_remote is
  signal si_incoming_in  : t_slot_bus_in_array(0 to g_CONFIG.slot_count-1);
  signal si_incoming_out : t_slot_bus_out_array(0 to g_CONFIG.slot_count-1);
  signal si_outgoing_in  : t_slot_bus_in_array(0 to g_CONFIG.slot_count-1);
  signal si_outgoing_out : t_slot_bus_out_array(0 to g_CONFIG.slot_count-1);

  signal incoming_stat : t_slot_status_out_array(0 to g_CONFIG.slot_count-1);
  signal outgoing_stat : t_slot_status_out_array(0 to g_CONFIG.slot_count-1);

  signal out_req, out_grant : std_logic_vector(g_CONFIG.slot_count-1 downto 0);

  signal out_cfgs : t_rmq_outgoing_slot_config_array(0 to g_CONFIG.slot_count-1);

  signal p_header       : t_rmq_rx_header;
  signal p_header_valid : std_logic;

begin  -- arch

  U_SI_Wishbone_Slave : entity work.mt_mqueue_wishbone_slave
    generic map (
      g_WITH_GCR => TRUE,
      g_CONFIG   => g_CONFIG)
    port map (
      clk_i             => clk_i,
      rst_n_i           => rst_n_i,
      incoming_status_i => incoming_stat,
      outgoing_status_i => outgoing_stat,
      incoming_o        => si_incoming_in,
      incoming_i        => si_incoming_out,
      outgoing_o        => si_outgoing_in,
      outgoing_i        => si_outgoing_out,
      slave_i           => si_slave_i,
      slave_o           => si_slave_o);

  gen_slots : for i in 0 to g_CONFIG.slot_count-1 generate

    U_Out_SlotX : entity work.mt_rmq_outgoing_slot
      generic map (
        g_ENTRIES => g_CONFIG.slot_config(i).entries,
        g_WIDTH   => g_CONFIG.slot_config(i).width)
      port map (
        clk_i      => clk_i,
        rst_n_i    => rst_n_i,
        stat_o     => outgoing_stat(i),
        inb_i      => si_outgoing_in(i),
        inb_o      => si_outgoing_out(i),
        tx_req_o   => out_req(i),
        tx_grant_i => out_grant(i),
        src_o      => src_o(i),
        src_i      => src_i(i),
        config_o   => out_cfgs(i));

    U_In_SlotX : entity work.mt_rmq_incoming_slot
      generic map (
        g_ENTRIES => g_CONFIG.slot_config(i).entries,
        g_WIDTH   => g_CONFIG.slot_config(i).width)
      port map (
        clk_i            => clk_i,
        rst_n_i          => rst_n_i,
        stat_o           => incoming_stat(i),
        outb_i           => si_incoming_in(i),
        outb_o           => si_incoming_out(i),
        snk_i            => snk_i(i),
        snk_o            => snk_o(i),
        p_header_valid_i => p_header_valid,
        p_header_i       => p_header,
        out_discard_i    => '0');

    rmq_status_o (i) <= not incoming_stat(i).empty;

  end generate gen_slots;

end architecture arch;
